#include <allegro.h>
#include <assert.h>
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include "graphics.h"
#include "map.h"
#include "misc.h"
#include "objects.h"

void translate(int x, int y)
{ int i, j, k;
  static int map2[2][256][256];

  for(i=0; i<2; i++)
    for(j=0; j<256; j++)
      for(k=0; k<256; k++)
      { if(j>=0 && j<=255 && j+x>=0 && j+x<=255 &&
           k>=0 && k<=255 && k+y>=0 && k+y<=255)
          map2[i][j][k]=map[i][j+x][k+y];
        else
          map2[i][j][k]=0;
      }

  for(i=0; i<2; i++)
    for(j=0; j<256; j++)
      for(k=0; k<256; k++)
        map[i][j][k]=map2[i][j][k];

  for(i=0; i<nobjs; i++)
  { objlist[i]->x-=x*32*256;
    objlist[i]->y-=y*32*256;
  }
}

int getobject(int x, int y)
{ int i;

  for(i=0; i<nobjs; i++)
    if(x==objlist[i]->x/32/256 && y==objlist[i]->y/32/256)
      return i;

  return -1;
}

int getobjecttype(int x, int y)
{ int i;

  for(i=0; i<nobjs; i++)
    if(x==objlist[i]->x/32/256 && y==objlist[i]->y/32/256)
      return objlist[i]->set;

  return 0;
}

void addremoveobject(int x, int y, int n)
{ int i;
  object *o;

  for(i=0; i<nobjs; i++)
    if(x==objlist[i]->x/32/256 && y==objlist[i]->y/32/256)
    { objlist[i]->destroy=1;
      removeobjects();
      return;
    }

  o=createobject(n, x*32*256, y*32*256);
  if(o) objlist[nobjs++]=o;
}

void dispedit(int n, int editing)
{ switch(n)
  { case KEYS:
      printfxy(0, 48, "color(a,s):%d", ((akey*)(objlist[editing]))->color);
      break;
    case LOCKS:
      printfxy(0, 48, "color(a,s):%d", ((lock*)(objlist[editing]))->color);
      break;
    case PLATFORM:
      printfxy(0, 48, "xvel(a,s):%d", ((platform*)(objlist[editing]))->xv);
      printfxy(0, 64, "yvel(z,x):%d", ((platform*)(objlist[editing]))->yv);
      break;
  }
}

void applyedit(char c, int n, int editing)
{ switch(n)
  { case KEYS:
      if(c=='a') ((akey*)(objlist)[editing])->color--;
      if(c=='s') ((akey*)(objlist)[editing])->color++;
      if(((akey*)(objlist)[editing])->color<0)
        ((akey*)(objlist)[editing])->color=4;
      if(((akey*)(objlist)[editing])->color>4)
        ((akey*)(objlist)[editing])->color=0;
      ((akey*)(objlist)[editing])->frame=((akey*)(objlist)[editing])->color;
      break;
    case LOCKS:
      if(c=='a') ((lock*)(objlist)[editing])->color--;
      if(c=='s') ((lock*)(objlist)[editing])->color++;
      if(((lock*)(objlist)[editing])->color<0)
        ((lock*)(objlist)[editing])->color=4;
      if(((lock*)(objlist)[editing])->color>4)
        ((lock*)(objlist)[editing])->color=0;
      ((lock*)(objlist)[editing])->frame=((lock*)(objlist)[editing])->color;
      break;
    case PLATFORM:
      switch(c)
      { case 'a': ((platform*)(objlist)[editing])->xv-=64; break;
        case 's': ((platform*)(objlist)[editing])->xv+=64; break;
        case 'z': ((platform*)(objlist)[editing])->yv-=64; break; 
        case 'x': ((platform*)(objlist)[editing])->yv+=64; break; 
      }
      break;
  }
}

void sortobjects()
{ int i;

  debuglog("sorting object list...");
  for(i=0; i<nobjs; i++)
  { if(objlist[i]->set==PLAYER)
    { object *swap;
      swap=objlist[i];
      objlist[i]=objlist[0];
      objlist[0]=swap;
    }
    objlist[i]->self=i;
  }
  debuglog("done sorting!\n");
}

#define OBJLIMIT 11
#define TILELIMIT 14

void editloop(char *filename)
{ int cx=0, cy=0, o=0, n=0, i;
  char drawbackground=1, drawlayer0=1, drawlayer1=1, drawobjects=1;
  int editing=-1;
  char c, draw=0;
  static int sp[TILELIMIT];

  if(nobjs) cx=objlist[0]->x/256/32, cy=objlist[0]->y/256/32;

  do
  { int x=cx*32-320, y=cy*32-384/2;

    if(x<0) x=0;
    if(y<0) y=0;
    if(x+640>256*32-1) x=256*32-640;
    if(y+480>256*32-1) y=256*32-480;
    drawscreen(x, y, drawbackground, drawlayer0, drawlayer1);
    if(drawobjects)
      for(i=nobjs-1; i>=0; i--)
      { if(objlist[i]->set==PLAYER)
          d_sprite(vscreen, objlist[i]->set, 0, objlist[i]->x/256-x,
                                                objlist[i]->y/256-y);
        else
          objlist[i]->draw(x, y);
      }
    draw_sprite(vscreen, spritesets[0].sprites[1], cx*32-x, cy*32-y);
    if(n)
    { if(o==2) d_sprite(vscreen, n, 0, 639-32, 0);
      else draw_sprite(vscreen, tilesets[n].tiles[sp[n]], 639-32, 0);
    }
    printfxy(0, 0, "x:%d y:%d", cx, cy);
    if(o==2) printfxy(0, 16, "object:%d", n);
    else printfxy(0, 16, "%s:%d (%d)", o?"background":"foreground", n, sp[n]);
    if(draw) printfxy(0, 32, "draw mode");
    if(editing>-1)
    { printfxy(0, 32, "editing object %d", editing);
      dispedit(n, editing);
    }
    blit(vscreen, screen, 0, 0, 0, 0, 640, 480);

    c=getch();
    switch(c)
    { case 'H': if(cy) cy--; break;
      case 'P': if(cy<255) cy++; break;
      case 'K': if(cx) cx--; break;
      case 'M': if(cx<255) cx++; break;
      case '[':
        if(editing==-1) n--;
        break;
      case ']':
        if(editing==-1) n++;
        break;
      case ';':
        if(o<2)
        { sp[n]--;
          if(sp[n]<0) sp[n]=tilesets[n].ntiles-1;
        }
        break;
      case '\'':
        if(o<2)
        { sp[n]++;
          if(sp[n]>=tilesets[n].ntiles) sp[n]=0;
        }
        break;
      case 'q': drawbackground=!drawbackground; break;
      case 'w': drawlayer0=!drawlayer0; break;
      case 'e': drawlayer1=!drawlayer1; break;
      case 'r': drawobjects=!drawobjects; break;
      case 'd': if(o!=2) draw=!draw; break;
      case '8': translate(0, 1); break;
      case '4': translate(1, 0); break;
      case '2': translate(0, -1); break;
      case '6': translate(-1, 0); break;
      case '`':
      { sortobjects();
        writemapfile(filename);
        break;
      }
      case 'p':
        o=2;
        draw=0;
        { if(editing<0)
          { editing=getobject(cx, cy);
            n=getobjecttype(cx, cy);
          }
          else editing=-1;
        }
        break;
      case 'o':
        if(++o>2) o=0;
        if(o==2) draw=0;
        else editing=-1;
        break;
      case 13:
        if(o==2) n=getobjecttype(cx, cy);
        else
        { n=map[o][cx][cy]&255;
          sp[n]=map[o][cx][cy]>>8;
        }
        break;
      case ' ':
        if(editing>-1) editing=-1;
        else
        { if(o==2) addremoveobject(cx, cy, n);
          else map[o][cx][cy]=n+sp[n]*256;
        }
        break;
    }

    if(n<0) n=(o==2?OBJLIMIT:TILELIMIT)-1;
    if((o!=2 && n>=TILELIMIT) || (o==2 && n>=OBJLIMIT)) n=0;
    
    if(editing>-1)
    { objlist[editing]->x=cx*32*256;
      objlist[editing]->y=cy*32*256;
      applyedit(c, n, editing);
    }
    if(draw) map[o][cx][cy]=n+sp[n]*256;
  } while(c!=27);
}

void editor(char *mapfile)
{ debuglog("Editor: " __DATE__ " " __TIME__ "\n");

  loadsprites();
  loadbackground();
  loadtiles();
  loadtranstable();
  readmapfile(mapfile);

  set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0);
  vscreen=create_bitmap(640, 480);
  set_palette(palette);

  editloop(mapfile);

  set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
}
