#include <assert.h>
#include <stdio.h>
#include "map.h"
#include "misc.h"
#include "objects.h"

unsigned short map[2][256][256];
atile animatedtiles[1024];
int nanimatedtiles=0;

void readmap(FILE *f)
{ int i, j, k;
  unsigned short n=0;
  short c;

  for(i=0; i<2; i++)
    for(j=0; j<256; j++)
      for(k=0; k<256; k++)
      { if(n && --n)
        { map[i][j][k]=0;
          continue;
        }
        c=readword(f);
        if(c) map[i][j][k]=c;
        else n=readword(f);
      }
  debuglog("done\n");
}

void readmap(char *s)
{ FILE *f=fopen(s, "rb");
  readmap(f);
  fclose(f);
}

void writemap(FILE *f)
{ int i, j, k;
  unsigned short n=0;

  for(i=0; i<2; i++)
  { for(j=0; j<256; j++)
      for(k=0; k<256; k++)
      { if(n==65535)
        { writeword(n, f);
          n=0;
        }
        if(map[i][j][k])
        { if(n) writeword(n, f);
          writeword(map[i][j][k], f);
          n=0;
        }
        else
        { if(!n) writeword(n, f);
          n++;
        }
      }
    if(n)
    { writeword(n, f);
      n=0;
    }
  }
}

void writeanimatedtiles(FILE *f)
{ writelong(nanimatedtiles, f);
  fwrite(animatedtiles, sizeof(atile), nanimatedtiles, f);
}

void readanimatedtiles(FILE *f)
{ nanimatedtiles=readlong(f);
  fread(animatedtiles, sizeof(atile), nanimatedtiles, f);
}

void readmapfile(char *s)
{ FILE *f=fopen(s, "rb");

  if(!f) return;
  readmap(f);
  readminiobjects(f);
  fclose(f);

  debuglog("mapfile read!\n");
}

void writemapfile(char *s)
{ FILE *f;

  assert(f=fopen(s, "wb"));
  writemap(f);
  writeminiobjects(f);
  fclose(f);
}

void readstate(FILE *f)
{ readmap(f);
  readobjects(f);
  readanimatedtiles(f);
}

void readstatefile(char *s)
{ FILE *f;

  assert(f=fopen(s, "rb"));
  readstate(f);
  fclose(f);
}

void writestate(FILE *f)
{ writemap(f);
  writeobjects(f);
  writeanimatedtiles(f);
}

void writestatefile(char *s)
{ FILE *f;

  f=fopen(s, "wb");
  writestate(f);
  fclose(f);
}

void makeanimatedtile(int x, int y, int dir, int start, int end, int loop,
                      int delay, int edgedelay, int frac)
{ animatedtiles[nanimatedtiles].x=x;
  animatedtiles[nanimatedtiles].y=y;
  animatedtiles[nanimatedtiles].dir=dir;
  animatedtiles[nanimatedtiles].start=start;
  animatedtiles[nanimatedtiles].end=end;
  animatedtiles[nanimatedtiles].loop=loop;
  animatedtiles[nanimatedtiles].delay=delay;
  animatedtiles[nanimatedtiles].edgedelay=edgedelay;
  animatedtiles[nanimatedtiles].frac=frac;
  nanimatedtiles++;
}

void initanimatedtiles()
{ int i, j;

  for(i=0; i<256; i++)
    for(j=0; j<256; j++)
    { if((map[0][i][j]&255)==WATER && map[0][i][j]>>8)
      { int dir;
        if(i%8>4)
        { map[0][i][j]=WATER+(9-i%8)*256;
          dir=-1;
        }
        else
        { map[0][i][j]=WATER+(i%8)*256;
          dir=1;
        }
        makeanimatedtile(i, j, dir, WATER+256, WATER+4*256,
                         PINGPONG, 20, 20, 0);
      }
      if((map[0][i][j]&255)==BLINK_BRICK)
      { int frac=0;
        if(map[0][i][j]==BLINK_BRICK+256)
        { frac=50;
          map[0][i][j]=BLINK_BRICK;
        }
        else if(map[0][i][j]==BLINK_BRICK+512)
        { frac=0;
          map[0][i][j]=BLINK_BRICK+256;
        }
        else if(map[0][i][j]==BLINK_BRICK+768)
        { frac=0;
          map[0][i][j]=BLINK_BRICK+1024;
        }
        else if(map[0][i][j]==BLINK_BRICK+1024)
          frac=50;
        makeanimatedtile(i, j, 1, BLINK_BRICK, BLINK_BRICK+4*256, PINGPONG,
                         20, 100, frac);
      }
      if((map[0][i][j]&255)==ONE_WAY)
        makeanimatedtile(i, j, 1, ((map[0][i][j]>>8)/6*6)*256+ONE_WAY,
                                  ((map[0][i][j]>>8)/6*6+5)*256+ONE_WAY,
                         FORWARD, 20, 20, 0);
    }
  debuglog("Nanimatedtiles: %d...", nanimatedtiles);
}

void animatetile(atile *t)
{ t->frac++;
  if(map[0][t->x][t->y]!=t->start && map[0][t->x][t->y]!=t->end)
  { if(t->frac<t->delay) return;
  }
  else if(t->frac<t->edgedelay) return;
  if(t->loop==PINGPONG)
  { if(map[0][t->x][t->y]==t->end && t->dir==1) t->dir=-1;
    if(map[0][t->x][t->y]==t->start && t->dir==-1) t->dir=1;
  }
  else
  { if(map[0][t->x][t->y]==t->end && t->dir==1) map[0][t->x][t->y]=t->start;
    if(map[0][t->x][t->y]==t->start && t->dir==-1) map[0][t->x][t->y]=t->end;
  }
  map[0][t->x][t->y]+=t->dir*256;
  t->frac=0;
}
