#include <assert.h>
#include <math.h>
#include <string.h>
#include "input.h"
#include "map.h"
#include "misc.h"
#include "objects.h"

inline int collide(object *o, int x2, int y2, int xw2, int yw2)
{ if(o->x>x2+(xw2-4)*256 || o->x+(o->xw-4)*256<x2 ||
     o->y>y2+(yw2-3)*256 || o->y+(o->yw-3)*256<y2) return 0;
  if(o->y<y2) return -1;
  return 1;
}

#define PLAYEROBJ 0
#define playerobj ((player *)objlist[0])

object *objlist[1024];
int nobjs=0;

inline int blocks(int tile)
{ return tile && tilesets[tile&255].type[tile>>8]!=UNBLOCKING;
}

int blocked(int self, int x, int y, int w, int h)
{ int i, j;
  int bx=x, by=y;

  x/=256, y/=256;

  if(x<0 || y<0 || x+w>8191 || y+h>8191)
    return 1;

  for(i=(x+3)/32; i<=(x+w-3)/32; i++)
    for(j=(y+3)/32; j<=(y+h-3)/32; j++)
    { int m=map[0][i][j];
      if(blocks(m))
      { char type=tilesets[m&255].type[m>>8];
        if(type==ONEWAYLEFT &&
           ((movingobject *)(objlist[self]))->xv<0)
          continue;
        if(type==ONEWAYRIGHT &&
           ((movingobject *)(objlist[self]))->xv>0)
          continue;
        return map[0][i][j];
      }
    }
  for(i=0; i<nobjs; i++)
    if(collide(objlist[i], bx, by, w, h) &&
       objlist[i]->blocking(self, x, y, w, h))
      return -1-i;

  return 0;
}

void object::draw(int cx, int cy)
{ if(x/256-cx+xw<0 || x/256-cx-xw>639 || y/256-cy+yw<0 || y/256-cy-yw>479)
    return;

  d_sprite(vscreen, set, frame, x/256-cx, y/256-cy);
}

void object::dumpstate()
{ debuglog("object:\nframe:%d self:%d set:%d x:%d y:%d xw:%d yw:%d\n",
                     frame,   self,   set,   x,   y,   xw,   yw);
}

void movingobject::dumpstate()
{ object::dumpstate();
  debuglog("x:%d.%d y:%d.%d xw:%d yw:%d xv:%d yv:%d\n",
           x/256, x%256, y/256, y%256, xw, yw, xv, yv);
}

void movingobject::move()
{ if(xv>terminalv) xv=terminalv;
  if(xv<-terminalv) xv=-terminalv;
  if(yv>terminalv) yv=terminalv;
  if(yv<-terminalv) yv=-terminalv;

  if(blocked(self, x+xv, y, xw, yw))
  { if(xv>256) while(!blocked(self, x+256, y, xw, yw)) x+=256;
    if(xv<-256) while(!blocked(self, x-256, y, xw, yw)) x-=256;
    xv=-xv*elasticity/256;
  }
  else x+=xv;
  if(blocked(self, x, y+yv, xw, yw))
  { if(yv>256) while(!blocked(self, x, y+256, xw, yw)) y+=256;
    if(yv<-256) while(!blocked(self, x, y-256, xw, yw)) y-=256;
    yv=-yv*elasticity/256;
    if(yv>-256 && yv<256) yv=0;
  }
  else y+=yv;

  if(!blocked(self, x, y+256, xw, yw)) yv+=gravity;
}

void player::draw(int xo, int yo)
{ int xc=x/256-xo, yc=y/256-yo, i;

  if(dead) d_sprite(vscreen, PLAYER, 6, xc, yc);
  else if(yv<-64) d_sprite(vscreen, PLAYER, 3, xc, yc);
  else if(yv>64) d_sprite(vscreen, PLAYER, 5, xc, yc);
  else if(ducking) d_sprite(vscreen, PLAYER, 4, xc, yc);
  else if(facing) d_sprite(vscreen, PLAYER, feet/1024, xc, yc);
  else draw_sprite_h_flip(vscreen, spritesets[PLAYER].sprites[feet/1024],
                          xc, yc);
  
  for(i=0; i<3; i++)
    if(keys[i]) d_sprite(vscreen, KEYS, i, 640-32*(3-i), 0);

  printfxy(0, 0, "score:%d", score);
  printfxy(320, 0, "x:%d y:%d", x/256, y/256);
  printfxy(320, 16, "%2s %2s %2s %2s", keyheld(KEY_UP)?"up":"",
                                       keyheld(KEY_DOWN)?"dn":"",
                                       keyheld(KEY_LEFT)?"lt":"",
                                       keyheld(KEY_RIGHT)?"rt":"");
}

void player::advframe()
{ int movemul=8;
  int waterlogged=(map[0][(int)(x/256+16)/32][(int)(y/256+16)/32]&255)==WATER;

  move();

  if(!keyheld(KEY_UP) || dead)
  { if(yv<0 && jumping) yv+=gravity*2;
    else jumping=0;
  }

  ducking=0;
  if(blocked(self, x, y+256, xw, yw))
  { if(xv>0) xv-=8;
    if(xv<0) xv+=8;
    if(xv<8 && xv>-8) xv=0;
    if(xv<-512) xv=-512;
    if(xv>512) xv=512;
    if(keyheld(KEY_DOWN) && !dead) ducking=1;
    movemul=16;
    if(!waterlogged && !dead && keyheld(KEY_UP) && !jumping)
      yv=-512-abs(xv/4), jumping=1;
  }

  if(waterlogged && !dead)
  { yv-=keyheld(KEY_UP)*16-keyheld(KEY_DOWN)*16;
    if(++nextbubble==250)
    { object *o=createobject(BUBBLE, x+256*13, y);
      nextbubble=0;
      objlist[nobjs++]=o;
    }
  }
  terminalv=1024-waterlogged*768;

  if(!ducking)
  { if(keyheld(KEY_LEFT) && !dead) xv-=movemul, facing=0;
    if(keyheld(KEY_RIGHT) && !dead) xv+=movemul, facing=1;
  }

  if(xv>512) xv=512;
  if(xv<-512) xv=-512;

  feet+=xv/2;
  if(feet<0) feet=1024*3-1;
  if(feet>=1024*3) feet=0;
}

int player::blocking(int what, int x2, int y2, int xw2, int yw2)
{ return objlist[what]->set==PLATFORM;
}

int player::push(int pxv, int pyv, int xalign, int yalign)
{ if(blocked(self, x+pxv, y+pyv, xw, yw)) return 0;
  x+=pxv, y+=pyv;
  x=x/256*256+xalign;
  y=y/256*256+yalign;
  return 1;
}

int player::push(int pxv, int pyv)
{ if(blocked(self, x+pxv, y+pyv, xw, yw)) return 0;
  x+=pxv, y+=pyv;
  return 1;
}

void blob::draw(int xo, int yo)
{ frame=4-(x/2048)%5;
  if(crushage)
    d_sprite(vscreen, BLOB, 5+crushage, x/256-xo, y/256-yo);
  else if(frame>2)
    draw_sprite_h_flip(vscreen, spritesets[BLOB].sprites[3-(frame-2)],
                       x/256-xo, y/256-yo);
  else d_sprite(vscreen, BLOB, frame, x/256-xo, y/256-yo);
}

void blob::advframe()
{ if(collide(playerobj, x, y, xw, yw))
  { if(playerobj->y<y-28*256+(crushage+1)*256*8 &&
       playerobj->y>y-28*256+crushage*256*8)
      if(++crushage==4)
      { destroy=1;
        if(playerobj->yv>0) playerobj->yv=-playerobj->yv;
      }
    if(playerobj->y>y-32*256+(crushage+1)*256*8)
      playerobj->dead=1;
  }
  else crushage=0;
  if(!crushage)
  { move();
    if(yv<0) yv=0;
  }
}

int blob::blocking(int what, int x2, int y2, int xw2, int yw2)
{ return what!=self && what!=PLAYEROBJ;
}

void bubble::advframe()
{ xv=rand()%2*512-256;
  move();
  if(map[0][(int)(x/256+16)/32][(int)(y/256+28)/32]!=WATER || !yv)
    destroy=1;
}

void supporter::move()
{ int b, playeron=0;

  if(xv>terminalv) xv=terminalv;
  if(xv<-terminalv) xv=-terminalv;
  if(yv>terminalv) yv=terminalv;
  if(yv<-terminalv) yv=-terminalv;

  if(collide(playerobj, x, y-256, xw, yw-2)) /* supporting player? */
  { playeron=1;
    playerobj->push(xv, yv, x%256, y%256);   /* move him */
  }

  x+=xv;
  y+=yv;

  b=blocked(self, x, y, xw, yw);
  if(b && !(b<0 && objlist[-b-1]->push(xv, yv)))
  { x-=xv;
    y-=yv;
    if(xv>256) while(!blocked(self, x+256, y, xw, yw)) x+=256;
    if(xv<-256) while(!blocked(self, x-256, y, xw, yw)) x-=256;
    if(yv>256) while(!blocked(self, x, y+256, xw, yw)) y+=256;
    if(yv<-256) while(!blocked(self, x, y-256, xw, yw)) y-=256;
    xv=-xv*elasticity/256;
    yv=-yv*elasticity/256;
  }
}

int supporter::blocking(int what, int x2, int y2, int xw2, int yw2)
{ return what!=self;
}

void platform::advframe()
{ move();
}

void coin::advframe()
{ if(++sp>19)
  { sp=0;
    if(++frame>7) frame=0;
  }
  if(collide(playerobj, x, y, xw, yw))
  { destroy=1;
    playerobj->score+=2;
  }
}

void spike::advframe()
{ if(++sp>79) sp=0;
  
  frame=sp/20;
  if(frame==3) frame=1;

  if(collide(playerobj, x, y, xw, yw) && playerobj->yv>0)
    playerobj->dead=1;
}

void akey::advframe()
{ if(collide(playerobj, x, y, xw, yw))
  { destroy=1;
    playerobj->keys[color]=1;
  }
}

int lock::blocking(int what, int x2, int y2, int xw2, int yw2)
{ if(playerobj->keys[color] && what==PLAYEROBJ)
  { destroy=1;
    return 0;
  }

  return 1;
}

int breaker::blocking(int what, int x2, int y2, int xw2, int yw2)
{ if(y2<y/256 && ++breakage>127) destroy=1;
  frame=breakage/16;

  return y2<y/256+(8-breakage/16);
}

void springboard::advframe()
{ if(jumpage)
  { jumpage+=2;
    if(jumpage<25)
    { frame=jumpage/5;
      playerobj->y=y-32*256+jumpage/5*4*256;
    }
    else if(jumpage<50)
    { frame=(49-jumpage)/5;
      if(jumpage<41)
        playerobj->y=y-32*256+(49-jumpage)/5*4*256;
      else if(jumpage==41)
      { playerobj->y=y-32*256;
        playerobj->yv=-1024;
      }
    }
    else jumpage=0;
  }
}

int springboard::blocking(int what, int x2, int y2, int xw2, int yw2)
{ if(y2<y/256-24 && playerobj->yv>=0 && !jumpage &&
     what==PLAYEROBJ) jumpage=1;
  return what!=PLAYEROBJ || !jumpage;
}

void portcullis::advframe()
{ if(collide(playerobj, x, y, xw, yw))
  { if(!fallage)
    { fallage=1;
      frame=1;
    }
  }
  if((playerobj->x<x-32*256 || x+32*256<playerobj->x) &&
     fallage && fallage<30)
  { fallage++;
    frame=1+fallage/5;
  }
}

int portcullis::blocking(int what, int x2, int y2, int xw2, int yw2)
{ return fallage>1;
}

object *createobject(int n, int x, int y)
{ object *o=0;

  switch(n)
  { case PLAYER: o=new player(nobjs, x, y); break;
    case COIN: o=new coin(nobjs, x, y); break;
    case SPIKE: o=new spike(nobjs, x, y); break;
    case KEYS: o=new akey(nobjs, x, y, 0); break;
    case LOCKS: o=new lock(nobjs, x, y, 0); break;
    case BREAKER: o=new breaker(nobjs, x, y); break;
    case SPRINGBOARD: o=new springboard(nobjs, x, y); break;
    case PORTCULLIS: o=new portcullis(nobjs, x, y); break;
    case BLOB: o=new blob(nobjs, x, y); break;
    case PLATFORM: o=new platform(nobjs, x, y); break;
    case BUBBLE: o=new bubble(nobjs, x, y); break;
  }

  return o;
}

void createobjects()
{ int i, j;

  for(i=0; i<256; i++)
    for(j=0; j<256; j++)
      if(map[1][i][j]==PLAYER)
        objlist[nobjs++]=new player(0, i*32*256, j*32*256);

  for(i=0; i<256; i++)
    for(j=0; j<256; j++)
    { object *o;
      o=createobject(map[1][i][j], i*32*256, j*32*256);      
      if(o) objlist[nobjs++]=o;
    }
}

void removeobjects()
{ int i;

  for(i=0; i<nobjs; i++)
    while(i<nobjs && objlist[i]->destroy)
    { delete objlist[i];
      objlist[i]=objlist[--nobjs];
    }
}

void writeobjects(FILE *f)
{ int i;

  writelong(nobjs, f);
  for(i=0; i<nobjs; i++)
  { assert(objlist[i]->set);
    writelong(objlist[i]->set, f);
    objlist[i]->writeself(f);
  }
}

void readobjects(FILE *f)
{ int i;

  for(i=0; i<nobjs; i++) delete objlist[i];

  nobjs=readlong(f);
  for(i=0; i<nobjs; i++)
  { object *o=0;
    int n;
    n=readlong(f);
    assert(n);
    o=createobject(n, 0, 0);
    if(o)
    { o->readself(f);
      objlist[i]=o;
    }
  }
}

void writeminiobjects(FILE *f)
{ int i;

  debuglog("writing itsybitsy objects...");

  writelong(nobjs, f);
  for(i=0; i<nobjs; i++)
  { assert(objlist[i]->set);
    writechar(objlist[i]->set, f);
    writechar(objlist[i]->x/256/32, f);
    writechar(objlist[i]->y/256/32, f);
    switch(objlist[i]->set)
    { case KEYS:
        writechar(((akey*)(objlist)[i])->color, f);
        break;
      case LOCKS:
        writechar(((lock*)(objlist)[i])->color, f);
        break;
      case PLATFORM:
        writelong(((platform*)(objlist)[i])->xv, f);
        writelong(((platform*)(objlist)[i])->yv, f);
        break;
    }
  }

  debuglog("done\n");
}

void readminiobjects(FILE *f)
{ int i, num;
  static char taken[256][256];

  memset(taken, 0, 256*256);

  num=readlong(f);
  for(i=0; i<num; i++)
  { int n, x, y;
    object *o=0;
    n=readchar(f);
    x=readchar(f);
    y=readchar(f);
    assert(n);
    o=createobject(n, x*32*256, y*32*256);
    assert(o);
    switch(n)
    { case KEYS:
        ((akey*)o)->color=((akey*)o)->frame=readchar(f);
        break;
      case LOCKS:
        ((lock*)o)->color=((lock*)o)->frame=readchar(f);
        break;
      case PLATFORM:
        ((platform*)o)->xv=readlong(f);
        ((platform*)o)->yv=readlong(f);
        break;
    }
    if(taken[x][y]) delete o;
    else
    { objlist[nobjs++]=o;
      taken[x][y]=1;
    }
  }
}
