#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>
#include "misc.h"

long lumpsize;

struct
{ char name[32];
  char filename[13];
  long offset;
  long length;
} pudindex[1024];

int indexp=0;

void error(char *fmt, ...)
{ va_list args;

  va_start(args, fmt);
  vprintf(fmt, args);
  va_end(args);
  exit(1);
}

void initpudfile(char *fn)
{ FILE *f;
  char buf[4];

  fflush(stdout);
  assert(f=fopen(fn, "rb"));
  fread(buf, 4, 1, f);
  assert(!memcmp(buf, "IPUD", 4));
  fread(buf, 4, 1, f);
  fseek(f, *(long *)(buf), SEEK_SET);
  while(!feof(f))
  { fread(pudindex[indexp].name, 32, 1, f);
    fread(&(pudindex[indexp].offset), 4, 1, f);
    fread(&(pudindex[indexp].length), 4, 1, f);
    strcpy(pudindex[indexp].filename, fn);
    indexp++;
  }
  fclose(f);
  indexp--;
}

char *loadlump(char *s)
{ int i, t;
  char *lump, *lump2;

  for(i=0, t=-1; i<indexp; i++)
    if(!strcmp(pudindex[i].name, s)) t=i;
  if(t==-1)
    error("Couldn't find lump: \"%s\"\n", s);
  lump=(char *)malloc(pudindex[t].length);
  if(!lump)
    error("Couldn't allocate %li bytes for lump \"%s\"\n", pudindex[t].length, s);

  FILE *f=fopen(pudindex[t].filename, "rb");
  fseek(f, pudindex[t].offset, SEEK_SET);
  fread(lump, pudindex[t].length, 1, f);
  fclose(f);

  int outsize=decodelong((unsigned char *)lump);

  lump2=(char *)malloc(outsize);
  int err=uncompress((Byte *)lump2, (uLong *)(&outsize), (Byte *)(lump+4),
                     pudindex[t].length-4);
  assert(err==Z_OK);
  free(lump);

  return lump2;
}
