#ifndef _OBJECTS_H
#define _OBJECTS_H

#include <stdio.h>
#include "graphics.h"

enum
{ PLAYER=1,
  BLOB=2,
  SPIKE=3,
  KEYS=4,
  LOCKS=5,
  COIN=6,
  BREAKER=7,
  SPRINGBOARD=8,
  PORTCULLIS=9,
  PLATFORM=10,
  BUBBLE=11
};

class object
{ public:
    int frame;
    int set;
    int self;
    int x, y;
    int xw, yw;
    int destroy;
    object(int s=0, int xp=0, int yp=0, int xwp=32, int ywp=32)
    { self=s, x=xp, y=yp, xw=xwp, yw=ywp;
      frame=destroy=0;
    }
    virtual void advframe() { return; }
    virtual void draw(int, int);
    virtual int push(int, int) { return 0; }
    virtual int blocking(int, int, int, int, int) { return 0; }
    virtual void dumpstate();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class movingobject:public object
{ protected:
    int gravity;
    int terminalv;
    int elasticity;
  public:
    int xv, yv;
    movingobject(int s, int a, int b, int c, int d, int e)
    { self=s, xv=a, yv=b, elasticity=c, gravity=d, terminalv=e;
      frame=0;
    }
    virtual void move();
    virtual void dumpstate();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class player:public movingobject
{ private:
    int jumping, ducking, facing, feet, nextbubble;
  public:
    int score, keys[5], dead;
    player(int self, int xp=0, int yp=0):movingobject(self, 0, 0, 0, 8, 512)
    { x=xp, y=yp, xw=yw=32;
      frame=jumping=facing=feet=score=keys[0]=keys[1]=keys[2]=keys[3]=keys[4]=
      dead=nextbubble=0;
      set=PLAYER;
    }
    virtual void advframe();
    virtual void draw(int, int);
    virtual int blocking(int, int, int, int, int);
    virtual int push(int, int);
    virtual int push(int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class blob:public movingobject
{ private:
    int crushage;
  public:
    blob(int self, int xp, int yp):movingobject(self, 120, 0, 256, 8, 512)
    { x=xp, y=yp;
      set=BLOB;
      crushage=0;
    }
    virtual void advframe();
    virtual void draw(int xo, int yo);
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class bubble:public movingobject
{ public:
    bubble(int self, int xp, int yp):movingobject(self, 0, -64, 0, 0, 64)
    { x=xp, y=yp;
      xw=7, yw=7;
      set=BUBBLE;
    }
    virtual void advframe();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class supporter:public movingobject
{ public:
    supporter(int s, int a, int b, int c, int d, int e)
      :movingobject(s, a, b, c, d, e){}
    virtual void move();
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class platform:public supporter
{ public:
    platform(int self, int xp, int yp):supporter(self, 128, 0, 256, 0, 128)
    { x=xp, y=yp;
      xw=32, yw=6;
      set=PLATFORM;
    }
    virtual void advframe();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class coin:public object
{ private:
    int sp;
  public:
    coin(int self, int x=0, int y=0):object(self, x, y, 32, 32)
    { sp=0;
      set=COIN;
    }
    virtual void advframe();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class spike:public object
{ private:
    int sp;
  public:
    spike(int self, int x=0, int y=0):object(self, x, y, 32, 32)
    { sp=0;
      set=SPIKE;
    }
    virtual void advframe();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class akey:public object
{ public:
    int color;
    akey(int self, int x, int y, int c):object(self, x, y, 32, 32)
    { frame=color=c;
      set=KEYS;
    }
    virtual void advframe();

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class lock:public object
{ public:
    int color;
    lock(int self, int x, int y, int c):object(self, x, y, 32, 32)
    { frame=color=c;
      set=LOCKS;
    }
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class breaker:public object
{ private:
    int breakage;
  public:
    breaker(int self, int x, int y):object(self, x, y, 32, 8)
    { breakage=0;
      set=BREAKER;
    }
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class springboard:public object
{ private:
    int jumpage;
  public:
    springboard(int self, int x, int y):object(self, x, y, 32, 32)
    { jumpage=0;
      set=SPRINGBOARD;
    }
    virtual void advframe();
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

class portcullis:public object
{ private:
    int fallage;
  public:
    portcullis(int self, int x, int y):object(self, x, y, 32, 32)
    { fallage=0;
      set=PORTCULLIS;
    }
    virtual void advframe();
    virtual int blocking(int, int, int, int, int);

    virtual void writeself(FILE *f);
    virtual void readself(FILE *f);
};

extern object *objlist[];
extern int nobjs;

extern object *createobject(int n, int x, int y);
extern void createobjects();
extern void removeobjects();
extern void writeobjects(FILE *f);
extern void readobjects(FILE *f);
extern void writeminiobjects(FILE *f);
extern void readminiobjects(FILE *f);

#endif
