#include <assert.h>
#include <stdio.h>

typedef struct
{ short height, width;
  unsigned char *image, *palette;
} pcx;

pcx loadpcx(char *filename)
{ FILE *f;
  int i, n=0;
  unsigned char c, count;
  pcx image;

  assert(f=fopen(filename, "rb"));

  fseek(f, 8, SEEK_SET);
  fread(&image.width, 2, 1, f);
  fread(&image.height, 2, 1, f);
  image.width++;
  image.height++;

  assert(image.image=(char *)malloc(image.width*image.height));

  fseek(f, 128, SEEK_SET);
  for(n=0; n<image.width*image.height;)
  { fread(&c, 1, 1, f);
    if(c<192) (image.image)[n++]=c;
    else
    { count=c-192;
      fread(&c, 1, 1, f);      
      for(; count>0; count--) image.image[n++]=c;
    }
  }

  assert(image.palette=(char *)malloc(768));
  fseek(f, -768, SEEK_END);
  fread(image.palette, 768, 1, f);

  fclose(f);

  return image;
}

void freepcx(pcx p)
{ if(p.palette) free(p.palette);
  if(p.image) free(p.image);
}

void writechar(char c, FILE *f)
{ fwrite(&c, 1, 1, f);
}

void writeword(int w, FILE *f)
{ writechar(w%256, f);
  writechar((w/256)%256, f);
}

void writelong(int l, FILE *f)
{ writeword(l%65536, f);
  writeword((l/65536)%65536, f);
}

int copyblocks(char *s, pcx p, FILE *f)
{ int x, y, w, h, n;
  int i, j;

  x=atoi(strtok(s, ","));
  y=atoi(strtok(0, ","));
  w=atoi(strtok(0, ","));
  h=atoi(strtok(0, ","));
  n=atoi(strtok(0, ","));
  printf("Copying %d blocks of %d, %d at %d, %d\n", n, w, h, x, y);
  assert(y+h<=p.height);

  for(i=0; i<n; i++)
  { assert(x+w<=p.width);
    writeword(w, f);
    writeword(h, f);
    for(j=0; j<h; j++)
      fwrite(p.image+(y+j)*p.width+x, w, 1, f);
    x+=w;
  }

  return n;
}

void stripcrlf(char *s)
{ for(; *s && *s!='\n'; s++);
  *s=0;
}

int main(int argc, char **argv)
{ FILE *f, *f2;
  pcx p={ 0, 0, 0, 0 };
  int n=0;

  assert(argc>=3);
  assert(f=fopen(argv[1], "r"));
  assert(f2=fopen(argv[2], "wb"));

  fwrite("ISPR\0\0\0\0", 8, 1, f2);

  while(!feof(f))
  { char s[256];

    fgets(s, 256, f);
    stripcrlf(s);
    printf("[%s]\n", s);
    if(isspace(*s)) n+=copyblocks(s, p, f2);
    else if(*s)
    { freepcx(p);
      p=loadpcx(s);
    }
  }

  fseek(f2, 4, SEEK_SET);
  writelong(n, f2);

  fclose(f);
  fclose(f2);

  return 0;
}
