#include <conio.h>
#include <allegro.h>
#include <assert.h>
#include <stdio.h>
#include <time.h>
#include "editor.h"
#include "input.h"
#include "graphics.h"
#include "map.h"
#include "misc.h"
#include "objects.h"
#include "pud.h"

char *mapfile=0;

#define TICKS_PER_SEC ticks_per_sec

int ticks_per_sec=120;

#define NFS 12

void resetgame()
{ int i;

  nanimatedtiles=0;

  for(i=0; i<nobjs; i++) objlist[i]->destroy=1;
  removeobjects();
  debuglog("nobjs==%d (should be 0)\n", nobjs);

  readmapfile(mapfile);
  initanimatedtiles();

  stopdemo();
  startdemo();
}

void loadgame(char *s)
{ int i;

  nanimatedtiles=0;
  for(i=0; i<nobjs; i++) objlist[i]->destroy=1;
  readstatefile(s);
}

void gameloop()
{ int x=0, y=0, i, j;
  int followobject=0;
  int fsk=0;
  static int fs[NFS]={ 1, 2, 3, 4, 5, 6, 8, 10, 12, 15, 20, 24 };
  int c, nc, failcount=0;

  do
  { c=uclock();
    x=objlist[followobject]->x/256-320;
    y=objlist[followobject]->y/256-240;
    if(x<0) x=0;
    if(y<0) y=0;
    if(x>256*32-640) x=256*32-640;
    if(y>256*32-480) y=256*32-480;
    drawscreen(x, y, 1, 1, 1);
    for(i=nobjs-1; i>=0; i--)
      objlist[i]->draw(x, y);
    for(i=0; i<fs[fsk]; i++)
    { refreshinput();
      for(j=0; j<nobjs; j++)
        objlist[j]->advframe();
      removeobjects();
      for(j=0; j<nanimatedtiles; j++)
        animatetile(animatedtiles+j);
    }
    printfxy(0, 16, "frameskip:%d", fs[fsk]);
    printfxy(0, 32, "fps:%d", TICKS_PER_SEC/fs[fsk]);
    printfxy(0, 48, "unreliability:%d", failcount);

    flush();
    nc=uclock();
    while(uclock()-c<UCLOCKS_PER_SEC/TICKS_PER_SEC*fs[fsk]);

    if(nc-c>UCLOCKS_PER_SEC/TICKS_PER_SEC*fs[fsk] && fsk<NFS-1)
    { if(++failcount>5) fsk++;
    }
    else
    { if(failcount) failcount--;
    }
    if(fsk && nc-c<UCLOCKS_PER_SEC/TICKS_PER_SEC*fs[fsk-1]) fsk--;

    if(key[KEY_S]) ticks_per_sec=30;
    if(key[KEY_N]) ticks_per_sec=120;
    if(key[KEY_F5]) writestatefile("save");
    if(key[KEY_F7]) loadgame("save");
    if(key[KEY_F9]) resetgame();
    if(key[KEY_SPACE])
    { ticks_per_sec=120;
    }
  } while(!key[KEY_ESC] && !keyheld(KEY_ESC));
}

int editmode=0;

void procparms(int argc, char **argv)
{ int i;

  debuglog("processing parameters...");
  for(i=1; i<argc; i++)
  { if(!memcmp(argv[i], "-r", 2))
    { debuglog("recording demo %s...", argv[i+1]);
      opendemo(argv[++i], RECORD);
    }
    else if(!memcmp(argv[i], "-p", 2))
    { debuglog("replaying demo %s...");
      opendemo(argv[++i], PLAYBACK);
    }
    else if(!memcmp(argv[i], "-e", 2)) 
    { debuglog("edit mode on...");
      editmode=1;
    }
    else
    { debuglog("map file selected...");
      mapfile=argv[i];
    }
  }

  debuglog("done\n");
}

int main(int argc, char **argv)
{ allegro_init();

#ifdef DEBUG
  debug=fopen("swarm.log", "a");
  fprintf(debug, "Startup: " __DATE__ " " __TIME__ "\n");
#endif

  procparms(argc, argv);
  if(!mapfile)
  { printf("Usage: %s [options] mapfile"
           "\noptions:"
           "\n\t-e[dit]"
           "\n\t-r[ecord] demoname"
           "\n\t-p[layback] demoname", argv[0]);
    return 1;
  }
  if(editmode && demostat)
  { printf("Demo recording in the editor not implemented.");
    return 2;
  }

  initpudfile("swarm.pud");

  debuglog("Loading sprites...");
  loadsprites();
  debuglog("done\n");
  debuglog("Loading background...");
  loadbackground();
  debuglog("done\n");
  debuglog("Loading tiles...");
  loadtiles();          
  debuglog("done\n");
  debuglog("Loading transparency table...");
  loadtranstable();
  debuglog("done\n");

  debuglog("Initializing graphics mode...");
  set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0);
  vscreen=create_bitmap(640, 480);
  set_palette(palette);  
  debuglog("done\n");

  if(editmode) editor(mapfile);
  else
  { debuglog("Initializing input...");
    install_keyboard();
    debuglog("done\n");
  
    debuglog("Loading map file...");
    readmapfile(mapfile);
    debuglog("done\n");

    debuglog("Initializing animated tiles...");
    initanimatedtiles();
    debuglog("done\n");

    assert(nobjs);
    startdemo();
    gameloop();
  }

  debuglog("Deinitializing graphics mode...");
  set_gfx_mode(GFX_TEXT, 80, 25, 0, 0);
  debuglog("done\n");

#ifdef DEBUG
  fclose(debug);
#endif
  stopdemo();

  return 0;
}
